﻿import {
	Color,
	FrontSide,
	LinearFilter,
	MathUtils,
	Matrix4,
	Mesh,
	PerspectiveCamera,
	Plane,
	RGBFormat,
	ShaderMaterial,
	UniformsLib,
	UniformsUtils,
	Vector3,
	Vector4,
	WebGLRenderTarget,
CustomBlending,
OneFactor,
SrcAlphaFactor,
DstAlphaFactor,
SrcColorFactor,
DstColorFactor,
AddEquation,
MaxEquation,
MinEquation,
OneMinusSrcAlphaFactor
} from '../build/three.module.js';


class WaterPlane extends Mesh {

	constructor( geometry, options = {} ) {

		super( geometry );

		const scope = this;

		const textureWidth = options.textureWidth !== undefined ? options.textureWidth : 512;
		const textureHeight = options.textureHeight !== undefined ? options.textureHeight : 512;

		const clipBias = options.clipBias !== undefined ? options.clipBias : 0.0;
		const alpha = options.alpha !== undefined ? options.alpha : 1.0;
		const time = options.time !== undefined ? options.time : 0.0;
		const normalSampler = options.waterNormals !== undefined ? options.waterNormals : null;
		const sunDirection = options.sunDirection !== undefined ? options.sunDirection : new Vector3( 0.70707, 0.70707, 0.0 );
		const sunColor = new Color( options.sunColor !== undefined ? options.sunColor : 0xffffff );
		const waterColor = new Color( options.waterColor !== undefined ? options.waterColor : 0x7F7F7F );
		const eye = options.eye !== undefined ? options.eye : new Vector3( 0, 0, 0 );
		const distortionScale = options.distortionScale !== undefined ? options.distortionScale : 20.0;
		const side = options.side !== undefined ? options.side : FrontSide;
		const fog = options.fog !== undefined ? options.fog : false;

//--//
		const basewaterColor = new Color( options.basewaterColor !== undefined ? options.basewaterColor : 0x7F7F7F );
		const basewaterAlpha = options.basewaterAlpha  !== undefined ? options.basewaterAlpha : 0.3;
		const mRefractionRatio = options.mRefractionRatio !== undefined ? options.mRefractionRatio : 1.02;
		const mFresnelBias = options.mFresnelBias !== undefined ? options.mFresnelBias : 0.1;
		const mFresnelPower = options.mFresnelPower !== undefined ? options.mFresnelPower : 2.0;
		const mFresnelScale = options.mFresnelScale !== undefined ? options.mFresnelScale : 1.0;
		const tCube = options.tCube !== undefined ? options.tCube : null;
//--//
		//

		const waterPlane = new Plane();
		const normal = new Vector3();
		const waterPlaneWorldPosition = new Vector3();
		const cameraWorldPosition = new Vector3();
		const rotationMatrix = new Matrix4();
		const lookAtPosition = new Vector3( 0, 0, - 1 );
		const clipPlane = new Vector4();

		const view = new Vector3();
		const target = new Vector3();
		const q = new Vector4();

		const textureMatrix = new Matrix4();

		const waterPlaneCamera = new PerspectiveCamera();

		const parameters = {
			minFilter: LinearFilter,
			magFilter: LinearFilter,
			format: RGBFormat
		};

		const renderTarget = new WebGLRenderTarget( textureWidth, textureHeight, parameters );

		if ( ! MathUtils.isPowerOfTwo( textureWidth ) || ! MathUtils.isPowerOfTwo( textureHeight ) ) {

			renderTarget.texture.generateMipmaps = false;

		}

		const waterPlaneShader = {

			uniforms: UniformsUtils.merge( [
				UniformsLib[ 'fog' ],
				UniformsLib[ 'lights' ],
				{
					'normalSampler': { value: null },
					'waterPlaneSampler': { value: null },
					'alpha': { value: 1.0 },
					'time': { value: 0.0 },
					'size': { value: 1.0 },
					'distortionScale': { value: 20.0 },
					'textureMatrix': { value: new Matrix4() },
					'sunColor': { value: new Color( 0x7F7F7F ) },
					'sunDirection': { value: new Vector3( 0.70707, 0.70707, 0 ) },
					'eye': { value: new Vector3() },
					'waterColor': { value: new Color( 0x555555 ) },
//--//
					'basewaterColor': { value: new Color( 0x555555 ) },
					'basewaterAlpha': { value: 0.3 },
					'mRefractionRatio': { value: 1.02 },
					'mFresnelBias': { value: 0.1 },
					'mFresnelPower': { value: 2.0 },
					'mFresnelScale': { value: 1.0 },
					'tCube': { value: null },
//--//
				}
			] ),

			vertexShader: /* glsl */`

				uniform mat4 textureMatrix;
				uniform float time;

				varying vec4 waterPlaneCoord;
				varying vec4 worldPosition;
//--//
				uniform float mRefractionRatio;
				uniform float mFresnelBias;
				uniform float mFresnelScale;
				uniform float mFresnelPower;

				varying vec3 vReflect;
				varying vec3 vRefract[3];
				varying float vReflectionFactor;
//--//

				#include <common>
				#include <fog_pars_vertex>
				#include <shadowmap_pars_vertex>
				#include <logdepthbuf_pars_vertex>

				void main() {


					waterPlaneCoord = modelMatrix * vec4( position, 1.0 );
					worldPosition = waterPlaneCoord.xyzw;
					waterPlaneCoord = textureMatrix * waterPlaneCoord;
					vec4 mvPosition =  modelViewMatrix * vec4( position, 1.0 );
					//gl_Position = projectionMatrix * mvPosition;

//--//
					//vec4 mvPosition = modelViewMatrix * vec4( position, 1.0 );
					//vec4 worldPosition = modelMatrix * vec4( position, 1.0 );

					vec3 worldNormal = normalize( mat3( modelMatrix[0].xyz, modelMatrix[1].xyz, modelMatrix[2].xyz ) * normal );

					vec3 I = worldPosition.xyz - cameraPosition;

					vReflect = reflect( I, worldNormal );
					vRefract[0] = refract( normalize( I ), worldNormal, mRefractionRatio );
					vRefract[1] = refract( normalize( I ), worldNormal, mRefractionRatio * 0.99 );
					vRefract[2] = refract( normalize( I ), worldNormal, mRefractionRatio * 0.98 );
					vReflectionFactor = mFresnelBias + mFresnelScale * pow( 1.0 + dot( normalize( I ), worldNormal ), mFresnelPower );

					gl_Position = projectionMatrix * mvPosition;

//--//

				#include <beginnormal_vertex>
				#include <defaultnormal_vertex>
				#include <logdepthbuf_vertex>
				#include <fog_vertex>
				#include <shadowmap_vertex>
			}`,

			fragmentShader: /* glsl */`



				uniform sampler2D waterPlaneSampler;
				uniform float alpha;
				uniform float time;
				uniform float size;
				uniform float distortionScale;
				uniform sampler2D normalSampler;
				uniform vec3 sunColor;
				uniform vec3 sunDirection;
				uniform vec3 eye;
				uniform vec3 waterColor;

				varying vec4 waterPlaneCoord;
				varying vec4 worldPosition;
//--//
				uniform samplerCube tCube;

				varying vec3 vReflect;
				varying vec3 vRefract[3];
				varying float vReflectionFactor;
				uniform vec3 basewaterColor;
				uniform float basewaterAlpha;
//--//

				vec4 getNoise( vec2 uv ) {
					vec2 uv0 = ( uv / 103.0 ) + vec2(time / 17.0, time / 29.0);
					vec2 uv1 = uv / 107.0-vec2( time / -19.0, time / 31.0 );
					vec2 uv2 = uv / vec2( 8907.0, 9803.0 ) + vec2( time / 101.0, time / 97.0 );
					vec2 uv3 = uv / vec2( 1091.0, 1027.0 ) - vec2( time / 109.0, time / -113.0 );
					vec4 noise = texture2D( normalSampler, uv0 ) +
						texture2D( normalSampler, uv1 ) +
						texture2D( normalSampler, uv2 ) +
						texture2D( normalSampler, uv3 );
					return noise * 0.5 - 1.0;
				}

				void sunLight( const vec3 surfaceNormal, const vec3 eyeDirection, float shiny, float spec, float diffuse, inout vec3 diffuseColor, inout vec3 specularColor ) {
					vec3 reflection = normalize( reflect( -sunDirection, surfaceNormal ) );
					float direction = max( 0.0, dot( eyeDirection, reflection ) );
					specularColor += pow( direction, shiny ) * sunColor * spec;
					diffuseColor += max( dot( sunDirection, surfaceNormal ), 0.0 ) * sunColor * diffuse;
				}

				#include <common>
				#include <packing>
				#include <bsdfs>
				#include <fog_pars_fragment>
				#include <logdepthbuf_pars_fragment>
				#include <lights_pars_begin>
				#include <shadowmap_pars_fragment>
				#include <shadowmask_pars_fragment>

				void main() {

					#include <logdepthbuf_fragment>

					vec4 noise = getNoise( worldPosition.xz * size );
					vec3 surfaceNormal = normalize( noise.xzy * vec3( 20.0, 1.0, 20.0 ) );

					vec3 diffuseLight = vec3(0.0);
					vec3 specularLight = vec3(0.0);

					vec3 worldToEye = eye-worldPosition.xyz;
					vec3 eyeDirection = normalize( worldToEye );
					sunLight( surfaceNormal, eyeDirection, 100.0, 2.0, 0.5, diffuseLight, specularLight );

					float distance = length(worldToEye);

					vec2 distortion = surfaceNormal.xz * ( 0.001 + 1.0 / distance ) * distortionScale;
					vec3 reflectionSample = vec3( texture2D( waterPlaneSampler, waterPlaneCoord.xy / waterPlaneCoord.w + distortion ) );

					float theta = max( dot( eyeDirection, surfaceNormal ), 0.0 );
					float rf0 = 0.3;
					float reflectance = rf0 + ( 1.0 - rf0 ) * pow( ( 1.0 - theta ), 5.0 );
					vec3 scatter = max( 0.0, dot( surfaceNormal, eyeDirection ) ) * waterColor;
					vec3 albedo = mix( ( sunColor * diffuseLight * 0.3 + scatter ) * getShadowMask(), ( vec3( 0.1 ) + reflectionSample * 0.9 + reflectionSample * specularLight ), reflectance);
					vec3 outgoingLight = albedo;
					vec4 wcol = vec4( outgoingLight, alpha );


//--//
					vec4 reflectedColor = textureCube( tCube, vec3( -vReflect.x, vReflect.yz ) );
					vec4 refractedColor = vec4( 1.0 );

					refractedColor.r = textureCube( tCube, vec3( -vRefract[0].x, vRefract[0].yz ) ).r;
					refractedColor.g = textureCube( tCube, vec3( -vRefract[1].x, vRefract[1].yz ) ).g;
					refractedColor.b = textureCube( tCube, vec3( -vRefract[2].x, vRefract[2].yz ) ).b;

					vec4 rcol = mix( refractedColor, reflectedColor, clamp( vReflectionFactor, 0.0, 1.0 ) );
					//gl_FragColor = mix( wcol, rcol, 0.8 );
					vec4 acol = mix( wcol, rcol, 0.8 );
					acol.w = clamp(acol.w + 0.6, 0.0, 1.0);
					//gl_FragColor = rcol;
					//gl_FragColor = wcol * rcol;
					//gl_FragColor = refractedColor;
                                        acol += vec4( basewaterColor * basewaterAlpha, 1.0);
					gl_FragColor = acol;
//--//






					#include <tonemapping_fragment>
					#include <fog_fragment>
				}`


		};


		const material = new ShaderMaterial( {
			fragmentShader: waterPlaneShader.fragmentShader,
			vertexShader: waterPlaneShader.vertexShader,
			uniforms: UniformsUtils.clone( waterPlaneShader.uniforms ),
			lights: true,
			side: side,
			fog: fog,
transparent: true,
opacity: 0.5,
blending: CustomBlending,
//blendSrc: OneFactor,
blendSrc: SrcAlphaFactor,
blendSrcAlpha: 0.5,
//blendEquation: AddEquation,
//blendEquation: MinEquation,
blendEquation: MaxEquation,
//blendEquationAlpha: 0.5,
//blendDst: DstColorFactor
//blendDst: DstAlphaFactor
blendDst: SrcAlphaFactor
//blendDst: OneFactor
//blendDst: OneMinusSrcAlphaFactor
//blendDstAlpha: 0.5

		} );

		material.uniforms[ 'waterPlaneSampler' ].value = renderTarget.texture;
		material.uniforms[ 'textureMatrix' ].value = textureMatrix;
		material.uniforms[ 'alpha' ].value = alpha;
		material.uniforms[ 'time' ].value = time;
		material.uniforms[ 'normalSampler' ].value = normalSampler;
		material.uniforms[ 'sunColor' ].value = sunColor;
		material.uniforms[ 'waterColor' ].value = waterColor;
		material.uniforms[ 'sunDirection' ].value = sunDirection;
		material.uniforms[ 'distortionScale' ].value = distortionScale;

		material.uniforms[ 'eye' ].value = eye;

//--//
		material.uniforms[ 'mRefractionRatio' ].value = mRefractionRatio;
		material.uniforms[ 'mFresnelBias' ].value = mFresnelBias;
		material.uniforms[ 'mFresnelPower' ].value = mFresnelPower;
		material.uniforms[ 'mFresnelScale' ].value = mFresnelScale;
		material.uniforms[ 'basewaterColor' ].value = basewaterColor;
		material.uniforms[ 'basewaterAlpha' ].value = basewaterAlpha;
		material.uniforms[ 'tCube' ].value = tCube;
//--//

		scope.material = material;

		scope.onBeforeRender = function ( renderer, scene, camera ) {

			waterPlaneWorldPosition.setFromMatrixPosition( scope.matrixWorld );
			cameraWorldPosition.setFromMatrixPosition( camera.matrixWorld );

			rotationMatrix.extractRotation( scope.matrixWorld );

			normal.set( 0, 0, 1 );
			normal.applyMatrix4( rotationMatrix );

			view.subVectors( waterPlaneWorldPosition, cameraWorldPosition );

			// Avoid rendering when mirror is facing away

			if ( view.dot( normal ) > 0 ) return;

			view.reflect( normal ).negate();
			view.add( waterPlaneWorldPosition );

			rotationMatrix.extractRotation( camera.matrixWorld );

			lookAtPosition.set( 0, 0, - 1 );
			lookAtPosition.applyMatrix4( rotationMatrix );
			lookAtPosition.add( cameraWorldPosition );

			target.subVectors( waterPlaneWorldPosition, lookAtPosition );
			target.reflect( normal ).negate();
			target.add( waterPlaneWorldPosition );

			waterPlaneCamera.position.copy( view );
			waterPlaneCamera.up.set( 0, 1, 0 );
			waterPlaneCamera.up.applyMatrix4( rotationMatrix );
			waterPlaneCamera.up.reflect( normal );
			waterPlaneCamera.lookAt( target );

			waterPlaneCamera.far = camera.far; // Used in WebGLBackground

			waterPlaneCamera.updateMatrixWorld();
			waterPlaneCamera.projectionMatrix.copy( camera.projectionMatrix );

			// Update the texture matrix
			textureMatrix.set(
				0.5, 0.0, 0.0, 0.5,
				0.0, 0.5, 0.0, 0.5,
				0.0, 0.0, 0.5, 0.5,
				0.0, 0.0, 0.0, 1.0
			);
			textureMatrix.multiply( waterPlaneCamera.projectionMatrix );
			textureMatrix.multiply( waterPlaneCamera.matrixWorldInverse );

			// Now update projection matrix with new clip plane, implementing code from: http://www.terathon.com/code/oblique.html
			// Paper explaining this technique: http://www.terathon.com/lengyel/Lengyel-Oblique.pdf
			waterPlane.setFromNormalAndCoplanarPoint( normal, waterPlaneWorldPosition );
			waterPlane.applyMatrix4( waterPlaneCamera.matrixWorldInverse );

			clipPlane.set( waterPlane.normal.x, waterPlane.normal.y, waterPlane.normal.z, waterPlane.constant );

			const projectionMatrix = waterPlaneCamera.projectionMatrix;

			q.x = ( Math.sign( clipPlane.x ) + projectionMatrix.elements[ 8 ] ) / projectionMatrix.elements[ 0 ];
			q.y = ( Math.sign( clipPlane.y ) + projectionMatrix.elements[ 9 ] ) / projectionMatrix.elements[ 5 ];
			q.z = - 1.0;
			q.w = ( 1.0 + projectionMatrix.elements[ 10 ] ) / projectionMatrix.elements[ 14 ];

			// Calculate the scaled plane vector
			clipPlane.multiplyScalar( 2.0 / clipPlane.dot( q ) );

			// Replacing the third row of the projection matrix
			projectionMatrix.elements[ 2 ] = clipPlane.x;
			projectionMatrix.elements[ 6 ] = clipPlane.y;
			projectionMatrix.elements[ 10 ] = clipPlane.z + 1.0 - clipBias;
			projectionMatrix.elements[ 14 ] = clipPlane.w;

			eye.setFromMatrixPosition( camera.matrixWorld );

			// Render

			const currentRenderTarget = renderer.getRenderTarget();

			const currentXrEnabled = renderer.xr.enabled;
			const currentShadowAutoUpdate = renderer.shadowMap.autoUpdate;

			scope.visible = false;

			renderer.xr.enabled = false; // Avoid camera modification and recursion
			renderer.shadowMap.autoUpdate = false; // Avoid re-computing shadows

			renderer.setRenderTarget( renderTarget );

			renderer.state.buffers.depth.setMask( true ); // make sure the depth buffer is writable so it can be properly cleared, see #18897

			if ( renderer.autoClear === false ) renderer.clear();
			renderer.render( scene, waterPlaneCamera );

			scope.visible = true;

			renderer.xr.enabled = currentXrEnabled;
			renderer.shadowMap.autoUpdate = currentShadowAutoUpdate;

			renderer.setRenderTarget( currentRenderTarget );

			// Restore viewport

			const viewport = camera.viewport;

			if ( viewport !== undefined ) {

				renderer.state.viewport( viewport );

			}

		};


	}

}

WaterPlane.prototype.isWaterPlane = true;

export { WaterPlane };
